/**
 * Created by chenrozenes and ofirhartuv on 05/11/2017.
 */
import Promise from 'bluebird';
import logger from '../logger';

const DEFAULT_INTERVAL = 1000;

/**
 * Base class that performs the pull logic from a queue
 * This class won't sync between different workers, and needs to be inherited in order to do so
 * (checkout PostgresPuller)
 */
class Puller {
  constructor(config = {}) {
    this.interval = config.interval || DEFAULT_INTERVAL;
    this.isRunning = false;
  }

  /**
   * Returns true if current worker can pull from queue
   */
  shouldPull() {
    return Promise.resolve(true);
  }

  /**
   * Pulls a message from a queue and calls the callback only pulled something
   * @param channel
   * @param queue
   * @param callback
   * @returns {*}
   */
  pull(channel, queue, callback) {
    return channel.get(queue, { noAck: false })
      .then((msg) => {
        // false means that there's no messages to pull in the queue
        if (msg !== false) {
          callback(msg);
        }
      });
  }

  /**
   * Pulling loop -
   * Checks if pulling is possible, pulls the message and tries again after waiting the interval
   * @param channel
   * @param queue
   * @param callback
   * @returns {Promise.<T>|*}
   */
  tryPull(channel, queue, callback) {
    if (!this.isRunning) return Promise.resolve();

    return this.shouldPull()
      .then((shouldPull) => {
        if (shouldPull) return this.pull(channel, queue, callback);
        return Promise.resolve();
      })
      .then(() => Promise.delay(this.interval))
      .then(() => {
        this.tryPull(channel, queue, callback);
      })
      .catch((error) => {
        logger.instance.error('[AMQP] Error occurred when trying to pull. ' +
          'continuing on the next interval', { error });

        throw error;
      });
  }

  /**
   * Starts Pulling loop
   * @param channel
   * @param queue
   * @param callback
   * @returns {Promise.<T>|*}
   */
  start(channel, queue, callback) {
    this.channel = channel;
    this.queue = queue;

    this.isRunning = true;
    logger.instance.debug(`[AMQP] Start pulling from queue ${queue}`);
    return this.tryPull(channel, queue, callback);
  }

  /**
   * Stops the pulling loop
   */
  stop() {
    logger.instance.debug('[AMQP] Stopping pulling');
    this.isRunning = false;
    // Wait for loop to stop
    return Promise.delay(this.interval * 2);
  }
}

export default Puller;
