/**
 * Created by asafdavid on 27/02/2017.
 */
import chai from 'chai';
import chaiAsPromised from 'chai-as-promised';
import sinonChai from 'sinon-chai';
import _ from 'lodash';
import * as Errio from 'errio';

import Router from '../../src/lib/handlers/router';
import Client from '../../src/lib/client';
import RabbitService from '../../src/service/rabbitService';
import TwilioService from './twilioService';
import BaseError from './baseError';
import TwilioServiceClassErrors from './twilioServiceClassErrors';
import * as inheritedServices from './twilioServicesWithInheritance';
import ExtendableError from '../../src/lib/error/extendableError';

chai.use(sinonChai);
chai.use(chaiAsPromised);
const expect = chai.expect;
const serverMeta = { serverExchangeName: 'temp' };
const domain = 'DEFAULT_DOMAIN';

class MyError extends ExtendableError {}
class MyNewError extends BaseError {}

describe('RabbitService', function () {
  describe('toRabbitHandlers', function () {
    it('returns a valid routes object', function () {
      const handlers = RabbitService.toRabbitHandlers(TwilioService.prototype._tasks);
      expect(handlers).to.be.an.array;
      expect(handlers).to.have.length.of(3);
      const send = _.find(handlers, { name: 'sendSms' });
      expect(send).to.have.all.keys('name', 'pattern', 'handle', 'options', 'middleware');
    });

    it('map params to the method', function (done) {
      const handlers = RabbitService.toRabbitHandlers(TwilioService.prototype._tasks);
      const send = _.find(handlers, { name: 'sendSms' });

      expect(send.handle(({ content: { number: '911', text: 'Help' } })))
        .to.eventually.be.equal('message sent Help to 911').notify(done);
    });
  });

  describe('register', function () {
    it('verifies that a router was provided', function () {
      expect(() => RabbitService.register()).to.throw(/router has to be an instance of Router/);
      expect(() => RabbitService.register({}, TwilioService))
        .to.throw(/router has to be an instance of Router/);
    });

    it('verifies that a service was provided', function () {
      const router = new Router();
      expect(() => RabbitService.register(router))
        .to.throw(/service has to be an instance of BaseService/);
      expect(() => RabbitService.register(router, {}))
        .to.throw(/service has to be an instance of BaseService/);
    });

    it('registers all of the handlers', function () {
      const router = new Router();
      RabbitService.register(router, TwilioService);
      expect(router.handlers).to.be.an.array;
      expect(router.handlers).to.have.length.of(3);
    });

    it('registers all only a subset of the handlers', function () {
      const router = new Router();
      RabbitService.register(router, TwilioService, ['sendSms']);
      expect(router.handlers).to.be.an.array;
      expect(router.handlers).to.have.length.of(1);
    });

    it('ignores unknown tasks', function () {
      const router = new Router();
      RabbitService.register(router, TwilioService, ['unknown']);
      expect(router.handlers).to.be.an.array;
      expect(router.handlers).to.have.length.of(0);
    });

    it('accepts an instance of a service as well', function () {
      const router = new Router();
      RabbitService.register(router, new TwilioService());
      expect(router.handlers).to.be.an.array;
      expect(router.handlers).to.have.length.of(3);
    });
  });

  describe('getClient', function () {
    const tests = [
      {
        name: 'with no service inheritance',
        service: TwilioService,
        serviceWithClassErrors: TwilioServiceClassErrors,
      },
      {
        name: 'with service inheritance',
        service: inheritedServices.TwilioService,
        serviceWithClassErrors: inheritedServices.TwilioServiceClassErrors,
      },
    ];

    tests.forEach(test =>
      describe(`${test.name}`, function () {
        describe('validation', function () {
          it('verifies that a service was provided', function () {
            expect(() => RabbitService.getClient())
            .to.throw(/service has to be an instance of BaseService/);
          });

          it('verifies that a client was provided', function () {
            expect(() => RabbitService.getClient(test.service))
            .to.throw(/client must be an instance of Client/);
          });

          it('verifies that a serverMeta was provided', function () {
            expect(() => RabbitService.getClient(test.service, new Client({ domain })))
            .to.throw(/wrong serverMeta was provided/);
            expect(() => RabbitService.getClient(test.service, new Client({ domain }), {}))
            .to.throw(/wrong serverMeta was provided/);
            expect(() => RabbitService.getClient(test.service, new Client({ domain }), { stam: '' }))
            .to.throw(/wrong serverMeta was provided/);
          });

          it('returns an interface object for a wanted service', function () {
            const client = RabbitService.getClient(
              test.service,
              new Client({ domain }),
              serverMeta
            );
            expect(client).to.be.an.object;
            expect(client).to.have.all.keys('sendSms', 'sendSmsWithBaseError', 'receiveSms');
          });
        });

        describe('sync', function () {
          beforeEach(function () {
            this.rabbitClient = new Client({ domain });
          });

          it('defines a sync task', function (done) {
            this.rabbitClient.sendSync = () =>
              Promise.resolve({ content: new Buffer('{"status": "ok"}') });

            /**
             * Defines a TwilioService client interface
             * @type {TwilioService}
             */
            const client = RabbitService.getClient(test.service, this.rabbitClient, serverMeta);
            expect(client.sendSms({ number: '911', text: 'Help' })).to.eventually
            .deep.equal({ status: 'ok' }).and.notify(done);
          });

          describe('handles rabbit errors correctly', function () {
            it('handles custom errors correctly', function (done) {
              this.rabbitClient.sendSync = () =>
                Promise.reject({
                  content: new Buffer(Errio.stringify(new MyError('my new error'))),
                });

              const client = RabbitService.getClient(test.service, this.rabbitClient, serverMeta);
              const rejection = new MyError();
              rejection.meta = { service: 'TwilioService', task: 'sendSms' };
              expect(client.sendSms({ number: '911', text: 'Help' }))
              .to.eventually.be.rejectedWith(MyError).and.notify(done);
            });

            it('handles Error correctly', function (done) {
              this.rabbitClient.sendSync = () =>
                Promise.reject({
                  content: new Buffer(Errio.stringify(new Error('my new error'))),
                });

              const client = RabbitService.getClient(test.service, this.rabbitClient, serverMeta);
              expect(client.sendSms('911', 'Help'))
              .to.eventually.be.rejectedWith(Error, 'my new error').and.notify(done);
            });

            it('handles unknown errors correctly', function (done) {
              this.rabbitClient.sendSync = () =>
                Promise.reject({ content: '{"name":"UnknownError","isOperational":true}' });

              const client = RabbitService.getClient(test.service, this.rabbitClient, serverMeta);
              expect(client.sendSms({ number: '911', text: 'Help' }))
              .to.eventually.be.rejectedWith(Error).and.notify(done);
            });

            it('stores the service name and task name in the error\'s metadata', function (done) {
              this.rabbitClient.sendSync = () =>
                Promise.reject({
                  content: new Buffer(Errio.stringify(new Error('my new error'))),
                });

              const client = RabbitService.getClient(test.service, this.rabbitClient, serverMeta);
              expect(client.sendSms({ number: '911', text: 'Help' }))
              .to.eventually.be.rejected
              .and.to.deep.equal({
                meta: {
                  service: 'TwilioService',
                  task: 'sendSms',
                },
                message: 'my new error',
                name: 'Error',
              })
              .and.notify(done);
            });

            describe('base error option', function () {
              it('wraps errors that are not instance of base error with the base error', function (done) {
                this.rabbitClient.sendSync = () =>
                  Promise.reject({
                    content: new Buffer(Errio.stringify(new Error('my new error'))),
                  });

                const client = RabbitService.getClient(test.service, this.rabbitClient, serverMeta);
                expect(client.sendSmsWithBaseError('911', 'Help'))
                .to.eventually.be.rejectedWith(BaseError, 'my new error').and.notify(done);
              });

              it('does not change errors that are instance of base error', function (done) {
                this.rabbitClient.sendSync = () =>
                  Promise.reject({
                    content: new Buffer(Errio.stringify(new MyNewError('my new error'))),

                  });

                const client = RabbitService.getClient(test.service, this.rabbitClient, serverMeta);
                expect(client.sendSmsWithBaseError({ number: '911', text: 'Help' }))
                .to.eventually.be.rejectedWith(MyNewError, 'my new error').and.notify(done);
              });
            });

            describe('class level errors', function () {
              it('wraps errors that are not instance of base error with the base error', function (done) {
                this.rabbitClient.sendSync = () =>
                  Promise.reject({
                    content: new Buffer(Errio.stringify(new Error('my new error'))),
                  });

                const client = RabbitService.getClient(test.serviceWithClassErrors,
                  this.rabbitClient,
                  serverMeta);
                expect(client.sendSms({ number: '911', text: 'Help' }))
                .to.eventually.be.rejectedWith(BaseError, 'my new error').and.notify(done);
              });
            });
          });
        });
      }));
  });

  describe('getGenericClient', function () {
    beforeEach(function () {
      this.rabbitClient = new Client({ domain });
    });

    it('defines a generic client', function (done) {
      const sendTopic = 'TWILIO.SEND';
      const msgToSend = { number: '911', text: 'Help' };
      const metaToSend = { domain: 'mydomain' };

      this.rabbitClient.sendSync = (server, topic, msg, meta) => {
        expect(topic).to.equal(sendTopic);
        expect(msg).to.deep.equal(msgToSend);
        expect(meta).to.deep.equal(metaToSend);
        return Promise.resolve({ content: new Buffer('{"status": "ok"}') });
      };

      const sync = true;
      const client = RabbitService.getGenericClient(
        'TwilioService',
        this.rabbitClient,
        sync,
        serverMeta
      );

      expect(client.dispatch(sendTopic, msgToSend, metaToSend)).to.eventually
      .deep.equal({ status: 'ok' }).and.notify(done);
    });

    it('sets the methods correctly using the mapping', function (done) {
      const mapping = {
        sendSms: 'TWILIO.SEND',
      };

      const msgToSend = { number: '911', text: 'Help' };
      this.rabbitClient.sendSync = (server, topic, msg) => {
        expect(topic).to.equal(mapping.sendSms);
        expect(msg).to.deep.equal(msgToSend);
        return Promise.resolve({ content: new Buffer('{"status": "ok"}') });
      };

      const sync = true;
      const client = RabbitService.getGenericClient(
        'TwilioService',
        this.rabbitClient,
        sync,
        serverMeta,
        mapping,
      );
      expect(client.sendSms(msgToSend)).to.eventually
      .deep.equal({ status: 'ok' }).and.notify(done);
    });
  });

  describe('async', function () {
    beforeEach(function () {
      this.rabbitClient = new Client({ domain });
    });

    it('defines an async task', function (done) {
      this.rabbitClient.sendAsync = () =>
        Promise.resolve();
      const context = {};

      /**
       * Defines a TwilioService client interface
       * @type {TwilioService}
       */
      const client = RabbitService.getClient(TwilioService, this.rabbitClient, serverMeta);
      expect(client.receiveSms(context)).to.eventually.be.fulfilled.and.notify(done);
    });

    it('handles rabbit errors correctly', function (done) {
      this.rabbitClient.sendSync = () =>
        Promise.reject();
      const context = {};

      /**
       * Defines a TwilioService client interface
       * @type {TwilioService}
       */
      const client = RabbitService.getClient(TwilioService, this.rabbitClient, serverMeta);
      expect(client.receiveSms(context))
        .to.eventually.be.rejectedWith(Error, 'could not send message to TWILIO.RECEIVE')
        .and.notify(done);
    });
  });
});
